from dns.resolver import NoAnswer  # noqa
from dns.exception import Timeout  # noqa


class BadAnswer(Exception):
    """On low level errors"""
    pass


class NoDNSSEC(Exception):
    """Whenever the recursive NS we use does not perform DNSSEC
    validation for our lookup, which might be for many reasons"""


class BadSignerName(Exception):
    pass


class DeleteDS(Exception):
    """Be explicit about CDNSKEY alg 0 when we have one."""


class BadCDNSKEY(Exception):
    pass


class NsLookupError(Exception):
    pass


class BadRcode(Exception):
    pass


# TPDA
class IncompleteURI(Exception):
    pass


class InvalidArgument(Exception):
    pass


class NoTPDA(Exception):
    """Whenever we're unable to lookup the DNSKEYs"""


class NoSignatureMatch(Exception):
    """No key matched the given signature"""


class DomainNotFound(Exception):
    pass


class TPDANotEnabled(Exception):
    pass


class ServiceNotPresent(Exception):
    pass


class UnsafeRedirect(Exception):
    pass



import dns
from . import exceptions


def rrset_rrsig(response):
    """Split and return rrsig and rrset from answer"""
    try:
        rrsig = response.find_rrset(response.answer, response.question[0].name,
                                    dns.rdataclass.IN, dns.rdatatype.RRSIG,
                                    response.question[0].rdtype)
        rrset = response.find_rrset(response.answer, response.question[0].name,
                                    dns.rdataclass.IN,
                                    response.question[0].rdtype)

        return rrset, rrsig
    except KeyError:
        return [], []


def signers(answer):
    """Takes a dns.resolver.Answer, and return signer
    names and key_tags. Check that signers are allowed
    to sign names.

        {'signer': [keytag, keytag], ..}

    """
    rrset, rrsig = rrset_rrsig(answer.response)
    qname = answer.response.question[0].name

    signer_names = set(sig.signer for sig in rrsig)
    if not all([qname.is_subdomain(sn) for sn in signer_names]):
        raise exceptions.BadSignerName

    ret = {}
    for sig in rrsig:
        ret.setdefault(sig.signer, []).append(sig.key_tag)

    return ret


def signed_by(answer, dnskey):
    """Checks that a given dns.resolver.Answer has been signed
    by the given dns Key object."""
    rrset, rrsig = rrset_rrsig(answer.response)

    for sig in rrsig:
        if sig.key_tag == dns.dnssec.key_id(dnskey):
            try:
                dns.dnssec.validate_rrsig(rrset, sig, {sig.signer: [dnskey]})
                return True
            except:
                pass
    return False


def trusted(answer):
    """Check if one signer in the dns.resolver.Answer is trusted"""
    from .resolver import query
    for signer in signers(answer).keys():
        keyans = query(signer, dns.rdatatype.DNSKEY, True)
        for key in keyans:
            if signed_by(answer, key):
                return True
    return False



from __future__ import absolute_import

import mock
import os
import unittest

from .. import tpda, exceptions


here = os.path.dirname(__file__)


class TestTPDA(unittest.TestCase):
    def setUp(self):
        self.client = tpda.Client('domain.com', '{}/test.key'.format(here))

    def _test_url_gen(self, fn, service, *args):
        with mock.patch('dnsknife.tpda.ServiceLocator.endpoint',
                        return_value='http://uri/') as mock_rdap:
            url = fn('tfz.net', *args)
            mock_rdap.assert_called_with('tfz.net', service)
            assert 'signature' in url
            assert 'source' in url
            assert 'expires' in url

        with (mock.patch('dnsknife.Checker.txt',
                         return_value=self.client.key_txt()),
              mock.patch('dnsknife.resolver.ns_for',
                         return_value=['1.2.3.4'])):
            trusted_url = tpda.validate_URI(url)
            assert 'source=domain.com' in trusted_url

        return url

    def test_nameserver_uri(self):
        uri = self._test_url_gen(self.client.nameservers_uri,
                                 'nameservers', ['ns1.', 'ns2.'])
        assert 'ns=ns1' in uri

    def test_dnssec_uri(self):
        uri = self._test_url_gen(self.client.dnssec_uri, 'dnssec',
                                 13, 'pouetpouet=', 257)
        assert 'flags=257' in uri
        assert 'pubkey=pou' in uri
        assert 'algorithm=13' in uri

    def test_website_uri(self):
        uri = self._test_url_gen(self.client.website_uri, 'website',
                                 ['1.2.3.4', '4.5.6.7'], None, 'alias.')
        assert 'cname=alias' in uri
        assert 'ipv4=' in uri

    def test_email_uri(self):
        uri = self._test_url_gen(self.client.email_uri, 'email',
                                 ['10 mx1.', '20 mx2.'])

        print uri
        assert 'mx=10+mx1' in uri

    def test_url_bad_signature(self):
        with mock.patch('dnsknife.tpda.ServiceLocator.endpoint',
                        return_value='http://uri/'):
            url = self.client.nameservers_uri('tfz.net', ['ns1.', 'ns2.'])

        with (mock.patch('dnsknife.Checker.txt',
                         return_value=self.client.key_txt()),
              mock.patch('dnsknife.resolver.ns_for',
                         return_value=['1.2.3.4'])):
            url = url.replace('source=domain', 'source=evil')
            self.assertRaises(exceptions.NoSignatureMatch,
                              tpda.validate_URI, url)

    def test_url_gen_override(self):
        client = tpda.Client('domain.com', '{}/test.key'.format(here),
                             override_uri='http://localhost:1234')
        url = client.nameservers_uri('tfz.net', ['ns1.', 'ns2.'])

        assert 'signature' in url
        assert 'source' in url
        assert 'expires' in url
        assert 'ns=' in url
        assert url.startswith('http://localhost:1234')



from __future__ import absolute_import, print_function

import base64
import datetime
import requests
from six.moves.urllib import parse

# TODO: improve key storage/publication
from Crypto.Signature import PKCS1_v1_5
from Crypto.PublicKey import RSA
from Crypto.Hash import SHA256

from . import Checker
from . import exceptions


def _qsl_get_one(qstring, param):
    """Return one unique param from query string"""
    args = parse.parse_qsl(qstring)
    sigvals = [arg[1] for arg in args if arg[0] == param]

    if len(sigvals) != 1:
        raise exceptions.InvalidArgument(param)

    return sigvals[0]


def validate_URI(uri):
    """Walk through an URI, lookup the set of DNSKEYs for the origin
    third party provider domain, validate URI signature against the
    found keys. If valid, returns the trustable URI - otherwise raise
    an exception.

    /!\ User MUST use the returned URI, as signature validation is
    only done on everything *before* the URI.
    """
    # Truncate the signature= part
    try:
        uri, sig = uri.split('&signature=')
        sig = parse.unquote(sig)
    except ValueError:
        raise exceptions.IncompleteURI

    pr = parse.urlparse(uri)
    if not pr.query:
        raise exceptions.IncompleteURI

    source = _qsl_get_one(pr.query, 'source')

    checker = Checker(source, dnssec=True)
    keys = [RSA.importKey(base64.b64decode(txt.encode('ascii')))
            for txt in checker.txt('_tpda').split('\n')]

    if not keys:
        raise exceptions.NoTPDA

    digest = SHA256.new()
    digest.update(uri.encode('ascii'))

    for key in keys:
        signer = PKCS1_v1_5.new(key)
        if signer.verify(digest, base64.b64decode(sig)):
            return uri

    raise exceptions.NoSignatureMatch


# Uses RDAP
class ServiceLocator:

    """Uses RDAP to find the registrar for a domain,
    registrar service URLs, or third party DNS operator
    service URLs"""

    def __init__(self, baseURI, timeout=5):
        self.baseURI = baseURI
        self.timeout = timeout

    def query(self, domain):
        uri = '{}/domain/{}'.format(self.baseURI, domain)
        r = requests.get(uri, timeout=self.timeout)
        if r.status_code == 404:
            raise exceptions.DomainNotFound()
        if r.status_code == 200:
            return r.json()
        return {}

    def endpoint(self, domain, service):
        d = self.query(domain)
        if not d.get('tpda_endpoints'):
            raise exceptions.TPDANotEnabled('no tpda endpoints on registrar')

        uri = d['tpda_endpoints'].get(service)
        if not uri:
            raise exceptions.ServiceNotPresent(
                'service is not part of tpda endpoints')

        return uri


# XXX Well.
STATIC_RDAP_SERVICES = [
    'http://rdap.gandi.net'
]


class Client:

    """
    TPDA Client for third party service providers.
    Example usage:

    (Generate "test.key" or import it from nameserver)

    from dnsknife import tpda
    tpda.Client('thirdparty.example.com','test.key').nameservers_uri(
               'client.example.com', ['ns1.example.com','ns2.example.com'])
    """

    def __init__(self, domain, key, lease_time=86400, override_uri=None):
        """Initialize a TPDA client. `domain` refers to the client
        domain where the _tpda DNSKEY record should be published.
        `key` is a local path to access this same private key.
        `lease_time` is the signature expiration delay"""
        self.domain = domain
        self.key = RSA.importKey(open(key).read())
        self.lease_time = lease_time
        self.override_uri = override_uri

    def key_txt(self):
        """Return text of key"""
        return base64.b64encode(self.key.publickey().exportKey('DER'))\
                     .decode('ascii')

    def txt_record_of_key(self):
        """Dumps RSA key as a TXT record"""
        txt = self.key_txt()
        strings = ['"{}"'.format(txt[i:i + 65])
                   for i in range(0, len(txt), 65)]
        return 'IN TXT {}'.format(" ".join(strings))

    def params_for_domain(self, domain):
        """Return the mandatory, common 3rd party parameters for this
        user's domain"""
        exp = (datetime.datetime.utcnow() + datetime.timedelta(
            seconds=self.lease_time)).strftime('%Y%m%d%H%M%S')

        return (('source', self.domain), ('domain', domain),
                ('expires', exp))

    def nameservers_uri(self, domain, ns_list):
        """Return the full signed URI to access this domain operator nameservers
        setup service, if available. XXX: Implement glue record/ipv4/ipv6
        management in here too.

        For now `ns_list` is simply a list of strings.
        """
        params = self.params_for_domain(domain)
        for ns in ns_list:
            params += ('ns', ns),

        return self.sign_for_domain(domain, 'nameservers', params)

    def record_uri(self, domain, records):
        """Return the full signed URI to access this domain operator record
        service, if available. records should be a tuple of (name, type, value)
        tuples such as:

            uri = domain.record_uri('example.com', (('@', 'A', '192.0.2.1'),
                ('www', 'CNAME', '@')))
        """

        params = self.params_for_domain(domain)
        for rname, rtype, rvalue in records:
            params += ('name', rname), ('type', rtype), ('value', rvalue)

        return self.sign_for_domain(domain, 'record', params)

    def dnssec_uri(self, domain, algorithm, pubkey, flags=257):
        """Return the full signed URI to access this domain operator dnssec
        first setup service and install given key, if available."""

        params = self.params_for_domain(domain)
        params += ('algorithm', algorithm), ('pubkey', pubkey), \
                  ('flags', flags)

        return self.sign_for_domain(domain, 'dnssec', params)

    def website_uri(self, domain, ipv4=None, ipv6=None, cname=None):
        """Return the full signed URI to change this domain web configuration
        as per spec. ipv4/v6 are used to setup zone apex A/AAAA records. CNAME
        can be used to setup 'WWW' alias."""

        params = self.params_for_domain(domain)
        for ipv4 in ipv4 or []:
            params += ('ipv4', ipv4),

        for ipv6 in ipv6 or []:
            params += ('ipv6', ipv6),

        if cname:
            params += ('cname', cname),

        return self.sign_for_domain(domain, 'website', params)

    def email_uri(self, domain, mx_list=[]):
        """Return the full signed URI to change this domain mail configuration
        as per spec. Format for each MX entry is "<priority> <exchange.>" as
        in MX records"""

        params = self.params_for_domain(domain)

        for m in mx_list:
            params += ('mx', m),

        return self.sign_for_domain(domain, 'email', params)

    def sign_for_domain(self, domain, service, params):
        def find(domain, service):
            for rdap in STATIC_RDAP_SERVICES:
                try:
                    return ServiceLocator(rdap).endpoint(domain, service)
                except Exception as e:
                    print(e)

        if not self.override_uri:
            base_url = find(domain, service)
        else:
            base_url = self.override_uri

        if not base_url:
            raise exceptions.NoTPDA

        uri = u'{}?{}'.format(base_url, parse.urlencode(params, True))

        signer = PKCS1_v1_5.new(self.key)
        digest = SHA256.new()
        digest.update(uri.encode('ascii'))

        sig = base64.b64encode(signer.sign(digest))
        return '{}&{}'.format(uri, parse.urlencode({'signature': sig}))



import dns.name
import dns.rdata
import dns.rdataclass
import dns.rdatatype
import dns.rdtypes.ANY.DNSKEY

dns.rdatatype.CDNSKEY = 60
dns.rdatatype._by_text['CDNSKEY'] = dns.rdatatype.CDNSKEY
dns.rdatatype._by_value[60] = 'CDNSKEY'

dns.rdata._rdata_modules[(dns.rdataclass.IN,
                          dns.rdatatype.CDNSKEY)] = dns.rdtypes.ANY.DNSKEY
dns.rdtypes.ANY.DNSKEY.CDNSKEY = dns.rdtypes.ANY.DNSKEY.DNSKEY

if not hasattr(dns.rdatatype, 'URI'):
    dns.rdatatype.URI = 256
    dns.rdatatype._by_text['URI'] = dns.rdatatype.URI

# For pickle/unpickle

del dns.name.Name.__setattr__



import sys
import os
import re

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'dnsknife/__init__.py')) as v_file:
    version = re.compile(r".*__version__ = '(.*?)'",
                         re.S).match(v_file.read()).group(1)

requires = [
    'requests',
    'pycrypto',
    'six'
]

if sys.version_info > (3,):
    requires.append('dnspython3')
else:
    requires.append('dnspython')

# commit 9f1c3988b64d4d95868825ecc48b00c1474bbf37 in PySocks
# introduced udp support. First available version is 1.5.6
pysocks = 'PySocks>=1.5.6'
extras_require = {
    'test': ['mock', pysocks],
    'socks': [pysocks]
}

tests_requires = requires + extras_require['test']

setup(
    name="dnsknife",
    version=version,
    packages=find_packages(),
    description="DNS tools",
    url="https://github.com/gandi/dnsknife",
    author="Gandi",
    author_email="feedback@gandi.net",
    license="GPLv3",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: Name Service (DNS)',
        'Intended Audience :: Developers',
    ],
    install_requires=requires,
    tests_require=tests_requires,
    test_suite='dnsknife.tests',
    extras_require=extras_require,
    keywords='dns lookup dnssec tpda rdap',
)
